@@ -15,7 +15,7 @@ import botocore
 import smtplib
 import ssl
 import requests
-from pandas.core.groupby import SeriesGroupBy, GroupBy
+# from pandas.core.groupby import SeriesGroupBy, GroupBy
 import orjson
 from time import sleep
 import traceback
@@ -25,7 +25,7 @@ from tethys_utils.data_models import Geometry, Dataset, DatasetBase, S3ObjectKey
 from geojson import Point
 import urllib3
 from multiprocessing.pool import ThreadPool, Pool
-from tethysts.utils import key_patterns, get_object_s3
+from tethysts.utils import key_patterns, get_object_s3, s3_connection, read_json_zstd, read_pkl_zstd
 from email.message import EmailMessage
 from datetime import date, datetime
 
@@ -52,39 +52,6 @@ agg_stat_mapping = {'mean': 'mean', 'cumulative': 'sum', 'continuous': None, 'ma
 ### Functions
 
 
-def s3_connection(conn_config, max_pool_connections=20):
-    """
-    Function to establish a connection with an S3 account. This can use the legacy connect (signature_version s3) and the curent version.
-
-    Parameters
-    ----------
-    conn_config : dict
-        A dictionary of the connection info necessary to establish an S3 connection.
-    max_pool_connections : int
-        The number of simultaneous connections for the S3 connection.
-
-    Returns
-    -------
-    S3 client object
-    """
-    s3_config = copy.deepcopy(conn_config)
-
-    if 'config' in s3_config:
-        config0 = s3_config.pop('config')
-        config0.update({'max_pool_connections': max_pool_connections})
-        config1 = boto3.session.Config(**config0)
-
-        s3_config1 = s3_config.copy()
-        s3_config1.update({'config': config1})
-
-        s3 = boto3.client(**s3_config1)
-    else:
-        s3_config.update({'config': botocore.config.Config(max_pool_connections=max_pool_connections)})
-        s3 = boto3.client(**s3_config)
-
-    return s3
-
-
 def list_objects_s3(s3_client, bucket, prefix, start_after='', delimiter='', continuation_token='', get_versions=False):
     """
     Wrapper S3 function around the list_objects_v2 base function with a Pandas DataFrame output.
@@ -203,20 +170,19 @@ def list_object_versions_s3(s3_client, bucket, prefix, next_key='', delimiter=No
     return f_df1
 
 
+# def get_last_date(s3_df, default_date='1900-01-01', date_type='date'):
+#     """
 
-def get_last_date(s3_df, default_date='1900-01-01', date_type='date'):
-    """
+#     """
+#     if not s3_df.empty:
+#         last_run_date = s3_df['KeyDate'].max()
+#     else:
+#         last_run_date = pd.Timestamp(default_date)
 
-    """
-    if not s3_df.empty:
-        last_run_date = s3_df['KeyDate'].max()
-    else:
-        last_run_date = pd.Timestamp(default_date)
+#     if date_type == 'str':
+#         last_run_date = last_run_date.strftime('%Y-%m-%d %H:%M:%S')
 
-    if date_type == 'str':
-        last_run_date = last_run_date.strftime('%Y-%m-%d %H:%M:%S')
-
-    return last_run_date
+#     return last_run_date
 
 
 def make_run_date_key(run_date=None):
@@ -259,26 +225,6 @@ def write_json_zstd(data, compress_level=1):
     return c_obj
 
 
-def read_json_zstd(obj):
-    """
-    Deserializer from a compressed zstandard json object to a dictionary.
-
-    Parameters
-    ----------
-    obj : bytes
-        The bytes object.
-
-    Returns
-    -------
-    Dict
-    """
-    dctx = zstd.ZstdDecompressor()
-    obj1 = dctx.decompress(obj)
-    dict1 = orjson.loads(obj1)
-
-    return dict1
-
-
 def write_pkl_zstd(obj, file_path=None, compress_level=1, pkl_protocol=pickle.HIGHEST_PROTOCOL):
     """
     Serializer using pickle and zstandard. Converts any object that can be pickled to a binary object then compresses it using zstandard. Optionally saves the object to disk. If obj is bytes, then it will only be compressed without pickling.
@@ -311,45 +257,15 @@ def write_pkl_zstd(obj, file_path=None, compress_level=1, pkl_protocol=pickle.HI
         return c_obj
 
 
-def read_pkl_zstd(obj, unpickle=True):
-    """
-    Deserializer from a pickled object compressed with zstandard.
+# def put_result_object_s3(s3, bucket, obj, dataset_id, station_id, run_date=None, content_type='application/zstd'):
+#     """
 
-    Parameters
-    ----------
-    obj : bytes or str
-        Either a bytes object that has been pickled and compressed or a str path to the file object.
-    unpickle : bool
-        Should the bytes object be unpickled or left as bytes?
+#     """
+#     run_date_key = make_run_date_key(run_date)
+#     ts_key = key_patterns['results'].format(dataset_id=dataset_id, station_id=station_id, run_date=run_date_key)
+#     s3.put_object(Body=obj, Bucket=bucket, Key=ts_key, ContentType=content_type, Metadata={'run_date': run_date_key})
 
-    Returns
-    -------
-    Python object
-    """
-    dctx = zstd.ZstdDecompressor()
-    if isinstance(obj, str):
-        with open(obj, 'rb') as p:
-            obj1 = dctx.decompress(p.read())
-    elif isinstance(obj, bytes):
-        obj1 = dctx.decompress(obj)
-    else:
-        raise TypeError('obj must either be a str path or a bytes object')
-
-    if unpickle:
-        obj1 = pickle.loads(obj1)
-
-    return obj1
-
-
-def put_result_object_s3(s3, bucket, obj, dataset_id, station_id, run_date=None, content_type='application/zstd'):
-    """
-
-    """
-    run_date_key = make_run_date_key(run_date)
-    ts_key = key_patterns['results'].format(dataset_id=dataset_id, station_id=station_id, run_date=run_date_key)
-    s3.put_object(Body=obj, Bucket=bucket, Key=ts_key, ContentType=content_type, Metadata={'run_date': run_date_key})
-
-    return ts_key
+#     return ts_key
 
 
 def results_data_integrety_checks(data, param_name, attrs, encoding, ancillary_variables=None):
@@ -403,21 +319,31 @@ def results_data_integrety_checks(data, param_name, attrs, encoding, ancillary_v
     return data
 
 
-# def station_data_integrety_checks(data, attrs=None, encoding=None):
-#     """
+def station_data_integrety_checks(data, attrs=None, encoding=None):
+    """
 
-#     """
-#     # Station data
-#     essentials = {'lat': (float, np.float), 'lon': (float, np.float), 'station_id': str, 'altitude': (int, float, np.float, np.int)}
+    """
+    # Station data
+    ignore_keys = ['station_id', 'lat', 'lon', 'name', 'altitude', 'ref', 'virtual_station', 'geometry']
+    other_keys = [s for s in data if s not in ignore_keys]
 
-#     for c in essentials:
-#         if not c in data.keys():
-#             raise ValueError('The station_data DataFrame must contain the field: ' + str(c))
+    if other_keys:
 
-#         if not isinstance(data[c], essentials[c]):
-#             raise ValueError('The station_data DataFrame field must have the data type(s): ' + str(essentials[c]))
+        if isinstance(attrs, dict):
+            for col in other_keys:
+                if not col in attrs:
+                    raise ValueError(col + ' must be in the attrs dict')
+        else:
+            raise TypeError('attrs must be a dict')
 
-#     return data
+        if isinstance(encoding, dict):
+            for col in other_keys:
+                if not col in encoding:
+                    raise ValueError(col + ' must be in the encoding dict')
+        else:
+            raise TypeError('encoding must be a dict')
+
+    return data
 
 
 def data_to_xarray(results_data, station_data, param_name, results_attrs, results_encoding, station_attrs=None, station_encoding=None, virtual_station=False, run_date=None, ancillary_variables=None, compression=False, compress_level=1):
@@ -447,26 +373,33 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
     Xarray Dataset or bytes object
 
     """
+    ## Station data prep
+    stn_m = StationBase(**station_data)
+    stn_data = orjson.loads(stn_m.json(exclude_none=True))
+    stn_data['lon'] = stn_data['geometry']['coordinates'][0]
+    stn_data['lat'] = stn_data['geometry']['coordinates'][1]
+    stn_data.pop('geometry')
+    if 'properties' in stn_data:
+        props = stn_data['properties'].copy()
+        for k, v in props.items():
+            stn_data[k] = v
+        stn_data.pop('properties')
+
     ## Integrity Checks
 
     # Time series data
     ts_data1 = results_data_integrety_checks(results_data, param_name, results_attrs, results_encoding, ancillary_variables)
 
     # Station data
-    stn_m = StationBase(**station_data)
-    stn_data = orjson.loads(stn_m.json(exclude_none=True))
-    stn_data['lon'] = stn_data['geometry']['coordinates'][0]
-    stn_data['lat'] = stn_data['geometry']['coordinates'][1]
-    stn_data.pop('geometry')
+    stn_data = station_data_integrety_checks(stn_data, attrs=station_attrs, encoding=station_encoding)
 
     ## Assign Attributes
-
     if isinstance(station_attrs, dict):
         attrs1 = copy.deepcopy(station_attrs)
     else:
         attrs1 = {}
 
-    attrs1.update({'station_id': {'cf_role': "timeseries_id", 'virtual_station': virtual_station}, 'lat': {'standard_name': "latitude", 'units': "degrees_north"}, 'lon': {'standard_name': "longitude", 'units': "degrees_east"}, 'altitude': {'standard_name': 'surface_altitude', 'long_name': 'height above the geoid to the lower boundary of the atmosphere', 'units': 'm'}})
+    attrs1.update({'station_id': {'cf_role': "timeseries_id", 'virtual_station': virtual_station}, 'virtual_station': {'long_name': 'Is this station a virtual or modeled station?'}, 'lat': {'standard_name': "latitude", 'units': "degrees_north"}, 'lon': {'standard_name': "longitude", 'units': "degrees_east"}, 'altitude': {'standard_name': 'surface_altitude', 'long_name': 'height above the geoid to the lower boundary of the atmosphere', 'units': 'm'}})
     if 'name' in stn_data.keys():
         attrs1.update({'name': {'long_name': 'station name'}})
     if 'ref' in stn_data.keys():
@@ -509,8 +442,9 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
         encoding1.update({'modified_date': {'_FillValue': -99999999, 'units': "days since 1970-01-01 00:00:00"}})
 
     ## Create the Xarray Dataset
-
     ds1 = results_data.to_xarray()
+
+    ## Assign the stn data to the main Dataset
     for k, v in stn_data.items():
         ds1[k] = v
 
@@ -609,7 +543,7 @@ def grp_ts_agg(df, grp_col, ts_col, freq_code, agg_fun, discrete=False, **kwargs
                 df3 = df2.reset_index().set_index(grp_col + [ts_col]).sort_index()
 
             else:
-                grp_col.extend([pd.Grouper(freq=freq_code, **kwargs)])
+                grp_col.extend([pd.Grouper(level=ts_col, freq=freq_code, **kwargs)])
                 df3 = df1.groupby(grp_col).agg(agg_fun)
 
             return df3
@@ -894,6 +828,7 @@ def tsreg(ts, freq=None, interp=False):
 #         raise ValueError('df should be either a Series or DataFrame.')
 #     return fun1
 
+
 def assign_ds_ids(datasets):
     """
     Parameters
@@ -1121,24 +1056,28 @@ def get_station_data_from_xr(data):
         ancillary_variables = attrs['ancillary_variables'].split(' ')
         data_vars.extend(ancillary_variables)
 
+    stn_fields = list(StationBase.schema()['properties'].keys())
+    stn_fields.extend(['lon', 'lat'])
+
     stn_vars = [v for v in vars1 if (not v in dims1) and (not v in data_vars)]
-    stn_data1 = {k: v['data'] for k, v in data[stn_vars].to_dict()['data_vars'].items()}
+    stn_data1 = {k: v['data'] for k, v in data[stn_vars].to_dict()['data_vars'].items() if k in stn_fields}
     stn_data1['geometry'] = create_geometry([stn_data1['lon'], stn_data1['lat']])
-    # stn_data1['lon'] = stn_data1['geometry']['coordinates'][0]
-    # stn_data1['lat'] = stn_data1['geometry']['coordinates'][1]
     stn_data1.pop('lon')
     stn_data1.pop('lat')
     stn_data1['altitude'] = round(stn_data1['altitude'], 3)
     if not 'virtual_station' in stn_data1:
         stn_data1['virtual_station'] = False
 
+    props = {s: {'data': data[s].to_dict()['data'], 'attrs': data[s].to_dict()['attrs']} for s in stn_vars if s not in stn_fields}
+    stn_data1['properties'] = props
+
     ## Check model
     stn_m = StationBase(**stn_data1)
 
     return stn_data1
 
 
-def process_station_summ(dataset_id, station_id, data, connection_config, bucket, mod_date=None):
+def process_station_summ(dataset_id, station_id, data, object_infos, mod_date=None):
     """
 
     """
@@ -1148,10 +1087,10 @@ def process_station_summ(dataset_id, station_id, data, connection_config, bucket
         mod_date = pd.Timestamp(mod_date).tz_localize(None)
 
     ## Determine the latest result
-    s3 = s3_connection(connection_config)
-    prefix = key_patterns['results'].split('{run_date}')[0].format(dataset_id=dataset_id, station_id=station_id)
+    # s3 = s3_connection(connection_config)
+    # prefix = key_patterns['results'].split('{run_date}')[0].format(dataset_id=dataset_id, station_id=station_id)
 
-    object_infos1 = process_object_keys(s3, bucket, prefix)
+    # object_infos1 = process_object_keys(s3, bucket, prefix)
     # obj_list = [s.dict() for s in object_infos1 if s.dict()['key'].find('results') > 0]
     # obj_keys_df = pd.DataFrame(obj_list)
     # last_run_date = obj_keys_df['run_date'].max()
@@ -1175,13 +1114,16 @@ def process_station_summ(dataset_id, station_id, data, connection_config, bucket
     # xr2 = xr.concat(data_list, dim='time', data_vars='minimal')
     # data = xr2.sel(time=~xr2.get_index('time').duplicated('last'))
 
+    ## Get the keys info list
+    # object_infos1 = info_dict[station_id]
+
     ## Generate the info for the recently created data
     station_id = str(data['station_id'].values)
     stats1 = get_new_stats(data)
 
     ## Put it all together
     stn_dict2 = get_station_data_from_xr(data)
-    stn_dict2.update({'dataset_id': dataset_id, 'stats': stats1, 'results_object_key': object_infos1, 'modified_date': mod_date})
+    stn_dict2.update({'dataset_id': dataset_id, 'stats': stats1, 'results_object_key': object_infos, 'modified_date': mod_date})
 
     station_m = Station(**stn_dict2)
 
@@ -1337,27 +1279,29 @@ def compare_datasets_from_s3(conn_config, bucket, new_data, add_old=False, read_
         base_key_pattern = key_patterns['results']
 
     ## Get list of keys
-    s3 = s3_connection(conn_config)
+    # s3 = s3_connection(conn_config)
 
     if isinstance(last_run_date_key, str):
         key_dict.update({'run_date': last_run_date_key})
         last_key = base_key_pattern.format(**key_dict)
-        try:
-            last_info1 = s3.head_object(Bucket=bucket, Key=last_key)
-            last_key1 = pd.DataFrame({'Key': [last_key]})
-        except:
-            last_key1 = pd.DataFrame()
+        # last_key1 = pd.DataFrame({'Key': [last_key]})
+        # try:
+        #     last_info1 = s3.head_object(Bucket=bucket, Key=last_key)
+        #     last_key1 = pd.DataFrame({'Key': [last_key]})
+        # except:
+        #     last_key1 = pd.DataFrame()
     else:
-        prefix_key = key_patterns['results'].split('{run_date}')[0].format(**key_dict)
-        all_keys = list_objects_s3(s3, bucket, prefix_key)
-        last_key1 = all_keys[all_keys['KeyDate'] == all_keys['KeyDate'].max()]
+        last_key = None
+        # prefix_key = key_patterns['results'].split('{run_date}')[0].format(**key_dict)
+        # all_keys = list_objects_s3(s3, bucket, prefix_key)
+        # last_key1 = all_keys[all_keys['KeyDate'] == all_keys['KeyDate'].max()]
 
     ## Get previous data and compare
     if isinstance(public_url, str):
         conn_config = public_url
 
-    if not last_key1.empty:
-        last_key = last_key1.iloc[0]['Key']
+    if isinstance(last_key, str):
+        # last_key = last_key1.iloc[0]['Key']
         p_old_one = get_object_s3(last_key, conn_config, bucket, 'zstd')
         xr_old_one = xr.load_dataset(p_old_one)
         xr_old_one['time'] = xr_old_one['time'].dt.round('s')
@@ -1371,88 +1315,88 @@ def compare_datasets_from_s3(conn_config, bucket, new_data, add_old=False, read_
 
 
 
-def get_filtered_obj_list(remote, dataset_list):
-    """
+# def get_filtered_obj_list(remote, dataset_list):
+#     """
 
-    """
-    base_prefix = key_patterns['results'].split('{dataset_id}')[0]
-    dataset_ids = [d['dataset_id'] for d in dataset_list]
+#     """
+#     base_prefix = key_patterns['results'].split('{dataset_id}')[0]
+#     dataset_ids = [d['dataset_id'] for d in dataset_list]
 
-    s3 = s3_connection(remote['connection_config'])
-    obj_df = list_objects_s3(s3, remote['bucket'], base_prefix)
+#     s3 = s3_connection(remote['connection_config'])
+#     obj_df = list_objects_s3(s3, remote['bucket'], base_prefix)
 
-    obj_df1 = obj_df[obj_df['KeyDate'].notnull()].copy()
-    obj_df1['dataset_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[2])
-    obj_df1['station_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[3])
-    obj_df1['result_type'] = 'results'
-    obj_df1.loc[obj_df1['Key'].str.contains('buffer'), 'result_type'] = 'buffer'
+#     obj_df1 = obj_df[obj_df['KeyDate'].notnull()].copy()
+#     obj_df1['dataset_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[2])
+#     obj_df1['station_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[3])
+#     obj_df1['result_type'] = 'results'
+#     obj_df1.loc[obj_df1['Key'].str.contains('buffer'), 'result_type'] = 'buffer'
 
-    obj_df2 = obj_df1[obj_df1['dataset_id'].isin(dataset_ids)].copy()
+#     obj_df2 = obj_df1[obj_df1['dataset_id'].isin(dataset_ids)].copy()
 
-    return obj_df2
+#     return obj_df2
 
 
-def get_last_results(obj_df):
-    """
+# def get_last_results(obj_df):
+#     """
 
-    """
-    last_date1 = obj_df.groupby(['dataset_id', 'station_id'])['KeyDate'].last().reset_index()
-    obj_df1 = pd.merge(last_date1, obj_df[['Key', 'dataset_id', 'station_id', 'KeyDate', 'result_type']], on=['dataset_id', 'station_id', 'KeyDate'])
+#     """
+#     last_date1 = obj_df.groupby(['dataset_id', 'station_id'])['KeyDate'].last().reset_index()
+#     obj_df1 = pd.merge(last_date1, obj_df[['Key', 'dataset_id', 'station_id', 'KeyDate', 'result_type']], on=['dataset_id', 'station_id', 'KeyDate'])
 
-    return obj_df1
+#     return obj_df1
 
 
-def filter_old_ones(obj_df, run_date, days_prior):
-    """
+# def filter_old_ones(obj_df, run_date, days_prior):
+#     """
 
-    """
-    if isinstance(run_date.tzname(), str):
-        run_date1 = run_date.tz_convert('UTC').tz_localize(None)
-    else:
-        run_date1 = run_date
+#     """
+#     if isinstance(run_date.tzname(), str):
+#         run_date1 = run_date.tz_convert('UTC').tz_localize(None)
+#     else:
+#         run_date1 = run_date
 
-    old_date = run_date1 - pd.DateOffset(days=days_prior)
+#     old_date = run_date1 - pd.DateOffset(days=days_prior)
 
-    obj_df1 = obj_df[obj_df['KeyDate'] < old_date].copy()
+#     obj_df1 = obj_df[obj_df['KeyDate'] < old_date].copy()
 
-    return obj_df1
+#     return obj_df1
 
 
-def process_buffer(row, remote, run_date_key):
-    """
+# def process_buffer(row, remote, run_date_key):
+#     """
 
-    """
-    data_list = []
-    for i, r in row.iterrows():
-        obj1 = get_object_s3(r['Key'], remote['connection_config'], remote['bucket'], 'zstd')
-        b1 = io.BytesIO(obj1)
-        xr1 = xr.load_dataset(b1)
-        data_list.append(xr1)
+#     """
+#     data_list = []
+#     for i, r in row.iterrows():
+#         obj1 = get_object_s3(r['Key'], remote['connection_config'], remote['bucket'], 'zstd')
+#         b1 = io.BytesIO(obj1)
+#         xr1 = xr.load_dataset(b1)
+#         data_list.append(xr1)
 
-    xr2 = xr.concat(data_list, dim='time', data_vars='minimal')
-    xr3 = xr2.sel(time=~xr2.get_index('time').duplicated('last'))
-    xr3.attrs.update({'history': run_date_key + ': Generated'})
+#     xr2 = xr.concat(data_list, dim='time', data_vars='minimal')
+#     xr3 = xr2.sel(time=~xr2.get_index('time').duplicated('last'))
+#     xr3.attrs.update({'history': run_date_key + ': Generated'})
 
-    key_dict = {'dataset_id': r['dataset_id'], 'station_id': r['station_id'], 'run_date': run_date_key}
-    new_key = key_patterns['results'].format(**key_dict)
+#     key_dict = {'dataset_id': r['dataset_id'], 'station_id': r['station_id'], 'run_date': run_date_key}
+#     new_key = key_patterns['results'].format(**key_dict)
 
-    cctx = zstd.ZstdCompressor(level=1)
-    c_obj = cctx.compress(xr3.to_netcdf())
+#     cctx = zstd.ZstdCompressor(level=1)
+#     c_obj = cctx.compress(xr3.to_netcdf())
 
-    s3 = s3_connection(remote['connection_config'])
+#     s3 = s3_connection(remote['connection_config'])
 
-    s3.put_object(Body=c_obj, Bucket=remote['bucket'], Key=new_key, ContentType='application/zstd', Metadata={'run_date': run_date_key})
+#     s3.put_object(Body=c_obj, Bucket=remote['bucket'], Key=new_key, ContentType='application/zstd', Metadata={'run_date': run_date_key})
 
 
-def process_buffer_threaded(obj_df, remote, run_date_key, threads=30):
-    """
+# def process_buffer_threaded(obj_df, remote, run_date_key, threads=30):
+#     """
 
-    """
-    grp1 = obj_df.groupby(['dataset_id', 'station_id'])
+#     """
+#     grp1 = obj_df.groupby(['dataset_id', 'station_id'])
 
-    input_list = [[row, remote, run_date_key] for i, row in grp1]
+#     input_list = [[row, remote, run_date_key] for i, row in grp1]
 
-    output = ThreadPool(threads).starmap(process_buffer, input_list)
+#     output = ThreadPool(threads).starmap(process_buffer, input_list)
 
 
 # def process_buffer_last_run_date(ex_dataset_id, dataset_list, run_date, remote, days_prior=7):
@@ -1550,22 +1494,22 @@ def process_run_date(processing_code, dataset_list, remote, run_date=None, save_
     return run_date_dict
 
 
-def process_buffer_results(remote, dataset_list):
-    """
-    Legacy function for processing buffer files into normal results files.
-    """
-    run_date1 = pd.Timestamp.today(tz='utc').tz_localize(None)
+# def process_buffer_results(remote, dataset_list):
+#     """
+#     Legacy function for processing buffer files into normal results files.
+#     """
+#     run_date1 = pd.Timestamp.today(tz='utc').tz_localize(None)
 
-    obj_df2 = get_filtered_obj_list(remote, dataset_list)
-    obj_df3 = get_last_results(obj_df2)
-    obj_df4 = filter_old_ones(obj_df3, run_date1, 1)
+#     obj_df2 = get_filtered_obj_list(remote, dataset_list)
+#     obj_df3 = get_last_results(obj_df2)
+#     obj_df4 = filter_old_ones(obj_df3, run_date1, 1)
 
-    last_run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
+#     last_run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
 
-    process_buffer_threaded(obj_df4, remote, last_run_date_key)
+#     process_buffer_threaded(obj_df4, remote, last_run_date_key)
 
 
-def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key, mod_date=None, sum_closed='right', other_closed='left', discrete=True, ts_local_tz=None, station_attrs=None, station_encoding=None):
+def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key, mod_date=None, sum_closed='right', other_closed='left', discrete=True, station_attrs=None, station_encoding=None):
     """
 
     """
@@ -1574,9 +1518,8 @@ def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key
         ancillary_variables = ['modified_date']
     else:
         ancillary_variables = []
-    #     mod_date = pd.Timestamp.today(tz='utc').round('s').tz_localize(None)
 
-    ts_data1 = data_df.copy()
+    tz_str = 'Etc/GMT{0:+}'
 
     ## Iterate through each dataset
     for ds in dataset_list:
@@ -1598,8 +1541,17 @@ def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key
         ## Pre-Process data
         qual_col = 'quality_code'
         freq_code = ds_mapping['frequency_interval']
+        utc_offset = ds_mapping['utc_offset']
         parameter = ds_mapping['parameter']
 
+        ts_data1 = data_df.copy()
+
+        # Convert times to local TZ if necessary
+        if (not freq_code in ['T', 'H', '1H']) and (not utc_offset == '0H'):
+            t1 = int(utc_offset.split('H')[0])
+            tz1 = tz_str.format(-t1)
+            ts_data1['time'] = ts_data1['time'].dt.tz_localize('UTC').dt.tz_convert(tz1).dt.tz_localize(None)
+
         ## Aggregate data if necessary
         # Parameter
         if freq_code == 'T':
@@ -1632,15 +1584,17 @@ def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key
 
         ## Convert to xarray
         df4 = df3.copy()
-        if isinstance(ts_local_tz, str):
-            df4['time'] = df4['time'].dt.tz_localize(ts_local_tz).dt.tz_convert('utc').dt.tz_localize(None)
+
+        # Convert time back to UTC if necessary
+        if (not freq_code in ['T', 'H', '1H']) and (not utc_offset == '0H'):
+            df4['time'] = df4['time'].dt.tz_localize(tz1).dt.tz_convert('utc').dt.tz_localize(None)
+
         df4.set_index(['time', 'height'], inplace=True)
 
-        new1 = data_to_xarray(df4, station_dict, parameter, attrs1, encoding1, run_date=run_date_key, ancillary_variables=ancillary_variables, compression='zstd')
+        new1 = data_to_xarray(df4, station_dict, parameter, attrs1, encoding1, station_attrs=station_attrs, station_encoding=station_encoding, run_date=run_date_key, ancillary_variables=ancillary_variables, compression='zstd')
 
         ## Update the data_dict
         ds_id = ds_mapping['dataset_id']
-        # stn_id = station_dict['station_id']
 
         data_dict[ds_id].append(new1)
 
@@ -1661,10 +1615,6 @@ def update_results_s3(processing_code, data_dict, run_date_dict, remote, threads
             The S3 bucket.
     threads : int
         The number of threads to use to process the data.
-    add_old : bool
-        Should the data in the S3 be added to the output? Ususally needed for writing to buffer with real-time data.
-    read_buffer : bool
-        Should the results buffer file be read instead of the normal results file?
     public_url : str
         Optional if there is a public URL to the object instead of using the S3 API directly.
 
@@ -1703,6 +1653,16 @@ def update_results_s3(processing_code, data_dict, run_date_dict, remote, threads
 
         run_date_key = run_date_dict[ds_id]
 
+        # Create the Key info dict
+        prefix = key_patterns['results'].split('{station_id}')[0].format(dataset_id=ds_id)
+
+        keys1 = list_objects_s3(s3, bucket, prefix)
+        obj_df1 = keys1[keys1.Key.str.contains('results.nc')].copy()
+        # obj_df1['dataset_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[2])
+        obj_df1['station_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[3])
+
+        last_date1 = obj_df1.groupby(['station_id'])['KeyDate'].last().reset_index()
+
 
         def update_result(result):
             """
@@ -1728,12 +1688,18 @@ def update_results_s3(processing_code, data_dict, run_date_dict, remote, threads
             if no_comparison:
                 up1 = new1
             else:
-                up1 = compare_datasets_from_s3(conn_config, bucket, new1, add_old=add_old, read_buffer=False, public_url=public_url)
+                last_date_key_df = last_date1[last_date1['station_id'] == stn_id]
+                if last_date_key_df.empty:
+                    last_date_key = None
+                else:
+                    last_date_key = make_run_date_key(last_date_key_df['KeyDate'].iloc[0])
+
+                up1 = compare_datasets_from_s3(conn_config, bucket, new1, add_old=add_old, read_buffer=False, last_run_date_key=last_date_key, public_url=public_url)
 
             ## Save results
             if isinstance(up1, xr.Dataset) and (len(up1[parameter].time) > 0):
 
-                print('Save results')
+                # print('Save results')
                 key_dict = {'dataset_id': ds_id, 'station_id': stn_id, 'run_date': run_date_key}
 
                 # if read_buffer:
@@ -1745,12 +1711,25 @@ def update_results_s3(processing_code, data_dict, run_date_dict, remote, threads
                 cctx = zstd.ZstdCompressor(level=1)
                 c_obj = cctx.compress(up1.to_netcdf())
 
-                s3.put_object(Body=c_obj, Bucket=bucket, Key=new_key, ContentType='application/zstd', Metadata={'run_date': mod_date_key})
+                obj_resp = s3.put_object(Body=c_obj, Bucket=bucket, Key=new_key, ContentType='application/zstd', Metadata={'run_date': mod_date_key})
 
                 ## Process stn data
-                print('Save station data')
+                # print('Save station data')
 
-                stn_m = process_station_summ(ds_id, stn_id, up1, conn_config, bucket, mod_date=mod_date_key)
+                ## Process object key infos
+                stn_obj_df1 = obj_df1[obj_df1['station_id'] == stn_id].drop(['LastModified', 'station_id'], axis=1).copy()
+                keys = stn_obj_df1['Key'].unique()
+                new_key_info = [new_key, obj_resp['ResponseMetadata']['HTTPHeaders']['etag'].replace('"', ''), len(c_obj), pd.Timestamp(run_date_key).tz_localize(None)]
+
+                if new_key in keys:
+                    stn_obj_df1[stn_obj_df1['Key'] == new_key] = new_key_info
+                else:
+                    stn_obj_df1.loc[len(stn_obj_df1)] = new_key_info
+
+                info1 = [S3ObjectKey(key=row['Key'], bucket=bucket, content_length=row['Size'], etag=row['ETag'], run_date=row['KeyDate']) for i, row in stn_obj_df1.iterrows()]
+
+                ## Final station processing and saving
+                stn_m = process_station_summ(ds_id, stn_id, up1, info1, mod_date=mod_date_key)
 
                 stn4 = orjson.loads(stn_m.json(exclude_none=True))
                 up_stns = put_remote_station(s3, bucket, stn4, run_date=mod_date_key)
@@ -1840,7 +1819,34 @@ def delete_result_objects_s3(conn_config, bucket, dataset_ids=None, keep_last=10
     return rem_keys
 
 
+def create_shifted_df(series, from_range, to_range, freq_code, agg_fun, ref_name, include_0=False, discrete=False, **kwargs):
+    """
 
+    """
+    if not isinstance(series, pd.Series):
+        raise TypeError('series must be a pandas Series.')
+    if not isinstance(series.index, pd.DatetimeIndex):
+        raise TypeError('The series index must be a pandas DatetimeIndex.')
+
+    df = series.reset_index()
+    data_col = df.columns[1]
+    ts_col = df.columns[0]
+    s2 = grp_ts_agg(df, None, ts_col, freq_code, agg_fun, discrete, **kwargs)[data_col]
+
+    if include_0:
+        f_hours = list(range(from_range-1, to_range+1))
+        f_hours[0] = 0
+    else:
+        f_hours = list(range(from_range, to_range+1))
+
+    df_list = []
+    for d in f_hours:
+        n1 = s2.shift(d, 'H')
+        n1.name = ref_name + '_' + str(d)
+        df_list.append(n1)
+    data = pd.concat(df_list, axis=1).dropna()
+
+    return data
 
 
 
